/***************************************************************
 * Name:      wxiscanimageprofile.h
 * Purpose:   ...
 * Author:    Daniel Nell (daniel.nell@nellresearch.de)
 * Created:   2009-01-12
 * Copyright: Daniel Nell (www.nellresearch.de)
 * License:   GPL
 **************************************************************/

#include "wx_pch.h"
#include <wx/config.h>

// Private headers
#include "wxiscanimageprofile.h"


/////////////////////////////////////////////////////////////////////////////
// Class wxIScanImageProfile
//
// Load configuration from the configuration object/file.
//
void wxIScanImageProfile::LoadConfig( int nProfileIndex, int nResolution )
{
    // Initialize members.
    m_nProfileIndex= nProfileIndex;

    // Get profile configuration.
    wxConfigBase *poConfig= wxConfigBase::Get();

    poConfig->SetRecordDefaults();
    poConfig->SetPath( wxString::Format( wxT( "/Profile/%d" ), m_nProfileIndex ) );

    m_strProfileName= poConfig->Read( wxT( "ProfileName" ), wxString::Format( wxT( "%d" ), m_nProfileIndex ) );
    x= poConfig->Read( wxT( "x" ), 0L );
    y= poConfig->Read( wxT( "y" ), 0L );
    w= poConfig->Read( wxT( "w" ), (long)( 210.0 / 25.4 * (double)nResolution ) );
    h= poConfig->Read( wxT( "h" ), (long)( 297.0 / 25.4 * (double)nResolution ) );
    bCrop= (bool)poConfig->Read( wxT( "Crop" ), (long)false );
    bGreyScale= (bool)poConfig->Read( wxT( "GreyScale" ), (long)false );
    bMonoChrome= (bool)poConfig->Read( wxT( "MonoChrome" ), (long)false );
    nMonoThreshold= poConfig->Read( wxT( "MonoThreshold" ), 128L );
    nRotate90Right= poConfig->Read( wxT( "Rotate90Right" ), 0L );
}

// Save configuration to the configuration object/file.
//
void wxIScanImageProfile::SaveConfig()
{
    // Get profile configuration.
    wxConfigBase *poConfig= wxConfigBase::Get();

    poConfig->SetRecordDefaults();
    poConfig->SetPath( wxString::Format( wxT( "/Profile/%d" ), m_nProfileIndex ) );

    poConfig->Write( wxT( "ProfileName" ), m_strProfileName );
    poConfig->Write( wxT( "x" ), (long)x );
    poConfig->Write( wxT( "y" ), (long)y );
    poConfig->Write( wxT( "w" ), (long)w );
    poConfig->Write( wxT( "h" ), (long)h );
    poConfig->Write( wxT( "Crop" ), (long)bCrop );
    poConfig->Write( wxT( "GreyScale" ), (long)bGreyScale );
    poConfig->Write( wxT( "MonoChrome" ), (long)bMonoChrome );
    poConfig->Write( wxT( "MonoThreshold" ), nMonoThreshold );
    poConfig->Write( wxT( "Rotate90Right" ), (long)nRotate90Right );
}

// Apply changes to the give image.
//
wxImage wxIScanImageProfile::ApplyToImage( wxImage oImage )
{
    // Crop image to the profile image rect.
    if( bCrop )
    {
        oImage.Resize( wxSize( w + x, h + y ), wxPoint( 0, 0 ), 255, 255, 255 );
        oImage= oImage.GetSubImage( wxRect( x, y, w, h ) );
    }

    // Convert the image to greyscale.
    if( bGreyScale )
    {
        oImage= oImage.ConvertToGreyscale();
    }

    // Convert the image to mono chrome.
    if( bMonoChrome )
    {
        oImage= ::Image2Mono( oImage, nMonoThreshold);
    }

    // Rotate the image 90° to the right (multiple).
    for( int i= 0; i < nRotate90Right; i++ )
    {
        oImage= oImage.Rotate90();
    }

    return oImage;
}


/////////////////////////////////////////////////////////////////////////////
// Global image manipulation functions
//
// Convert an image to monochrome (black and white).
//
wxImage Image2Mono( const wxImage &oSrcImage, int nMonoThreshold )
{
    wxImage oImage= oSrcImage.ConvertToGreyscale();

    int nCount= oImage.GetWidth() * oImage.GetHeight() * 3;
    unsigned char *pImageDataSrc= oImage.GetData();
    unsigned char *pImageDataDest= (unsigned char *)malloc( nCount * sizeof( unsigned char ) );

    if( !pImageDataDest )
    {
        ::wxLogError( wxString( wxT( "Image2Mono - " ) )
                      + _( "Cannot allocate memory for destination image." ) );
        return oImage;
    }

    // ...
    unsigned char *pIndex= pImageDataDest;

    for( int i= 0; i < nCount; i += 3 )
    {
        unsigned char nValue= ( (*pImageDataSrc) < nMonoThreshold ) ? 0 : 255;

        // Red.
        (*pIndex)= nValue;
        pImageDataSrc++;
        pIndex++;

        // Green.
        (*pIndex)= nValue;
        pImageDataSrc++;
        pIndex++;

        // Blue.
        (*pIndex)= nValue;
        pImageDataSrc++;
        pIndex++;
    }

    oImage.SetData( pImageDataDest );
    return oImage;
}

/// \brief Calculate the value of x^y.
//
inline static double x_exp_y( double x, double y )
{
    return exp( y * log( x ) );
}

/// \brief Limit the integer value to the intervall [0..255].
//
inline static int Limit255( int x )
{
    if( x > 255 )
    {
        return 255;
    }
    else if( x < 0 )
    {
        return 0;
    }
    else
    {
        return x;
    }
}

// Modify brightness, contrast, gamma value, red, green
// and blue channel of an image.
//
wxImage EnhenceColours( const wxImage &oSrcImage,
                        int nBrightness, int nContrast, int nGamma,
                        int nRed, int nGreen, int nBlue )
{
    wxImage oImage= oSrcImage;

    // ...
    int nCount= oImage.GetWidth() * oImage.GetHeight() * 3;
    unsigned char *pImageDataSrc= oImage.GetData();
    unsigned char *pImageDataDest= (unsigned char *)malloc( nCount * sizeof( unsigned char ) );

    if( !pImageDataDest )
    {
        ::wxLogError( wxString( wxT( "EnhenceColours - " ) )
                      + _( "Cannot allocate memory for destination image." ) );
        return oImage;
    }

    // ...
    int nIncRed=   nRed   + nBrightness;
    int nIncGreen= nGreen + nBrightness;
    int nIncBlue=  nBlue  + nBrightness;
    double nContrastFactor= exp( (double)nContrast / 75.0 );
    double nInvGamma= 1.0 / ( (double)nGamma / 100.0 );

    // Recalculate the image as a whole. That is calculate brightness,
    // contrast, gamma and color correction all at once.
    //
    // NOTE: We use floating point (double) calculation for not to cut off
    //       values as early as 8 bit integer calculation would do (for
    //       numeric stability). The disadvantage is that the transformation
    //       of one image of about 8 Mega pixels (e. g. an A4 page) even on
    //       a 1.6 GHertz machine lasts for more than 5s.
    //
    unsigned char *pIndex= pImageDataDest;
    for( int i= 0; i < nCount; i += 3 )
    {
        // Red.
        (*pIndex)= Limit255( (int)( x_exp_y( ( (double)( (*pImageDataSrc) + nIncRed   ) * nContrastFactor ) / 255.0,
                                             nInvGamma ) * 255.0 ) );
        pImageDataSrc++;
        pIndex++;

        // Green.
        (*pIndex)= Limit255( (int)( x_exp_y( ( (double)( (*pImageDataSrc) + nIncGreen ) * nContrastFactor ) / 255.0,
                                               nInvGamma ) * 255.0 ) );
        pImageDataSrc++;
        pIndex++;

        // Blue.
        (*pIndex)= Limit255( (int)( x_exp_y( ( (double)( (*pImageDataSrc) + nIncBlue  ) * nContrastFactor ) / 255.0,
                                               nInvGamma ) * 255.0 ) );
        pImageDataSrc++;
        pIndex++;
    }

    oImage.SetData( pImageDataDest );
    return oImage;
}
