/////////////////////////////////////////////////////////////////////////////
// Project:     wxIView
// Purpose:     Complex wxWidgets sample
// Name:        wxiviewcanvas.cpp
// Author:      Daniel Nell
// Created:     2005/10/18
// Copyright:   (c) Daniel Nell
// Licence:     wxWindows license
// Modified by:
/////////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation, includes "wx/wx.h".
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
   #include <wx/wx.h>
#endif

// Include additional wxWidgets headers.
#include <wx/image.h>
#include <wx/timer.h>

// Include private wxIView headers.
#include "wxiviewcanvas.h"


//////////////////////////////////////////////////////////
// The canvas window
//
// ..
//
DEFINE_EVENT_TYPE( wxEVT_IVIEWCANVASSIZE )

// Event table,
//
BEGIN_EVENT_TABLE( wxIViewCanvas, wxScrolledWindow )
    EVT_PAINT(           wxIViewCanvas::OnPaint  )
    EVT_SIZE(            wxIViewCanvas::OnSize   )
    EVT_TIMER( wxID_ANY, wxIViewCanvas::OnTimer  )
    EVT_MOUSE_EVENTS(    wxIViewCanvas::OnMouse_ )
END_EVENT_TABLE()


// Standard constructor.
//
wxIViewCanvas::wxIViewCanvas( wxWindow *opParent, wxWindowID nId,
                              const wxPoint &oPos, const wxSize &oSize,
                              long nStyle, const wxString& strName )
  : wxScrolledWindow( opParent, nId, oPos, oSize, nStyle, strName ),
    m_nOptions( wxICANVAS_MOUSE_SCROLL )
{
    // Initialization.
    m_oTimer.SetOwner( this );

    // Reset selection.
    ClearSelection();
}

// Virtual destructor.
//
wxIViewCanvas::~wxIViewCanvas()
{
    // Clean up.
    // [Nothing to do, yet.]
}

// Event handlers.
//
// Creates and prepares a device context and calls Paint( wxDC& oDc ).
//
void wxIViewCanvas::OnPaint( wxPaintEvent& oEvent )
{
    // Create a device context...
    //
    wxPaintDC oDc( this );

    // ... and prepare it even if there is not a bitmap to paint
    // for to repaint at least the window background etc.
    //
    PrepareDC( oDc );

    // Paint (the bitmap).
    Paint( oDc );
}

// Handle resizing.
//
void wxIViewCanvas::OnSize( wxSizeEvent& WXUNUSED( oEvent ) )
{
	// Check if the timer is running, and...
    if( m_oTimer.IsRunning() )
    {
        // ... stop it.
        m_oTimer.Stop();
    }

	// Create a one shot timer event in 50ms.
    m_oTimer.Start( WXIVIEWCANVASDELAY, wxTIMER_ONE_SHOT );
}

// Handle timer events.
//
void wxIViewCanvas::OnTimer( wxTimerEvent& WXUNUSED( oEvent ) )
{
    // Notify the next window in the parent chain that the canvas size
    // has changed.
    //
    // Note: An event of type wxIViewCanvasSizeEvent is propageted
    //       to the parent window chain by default.
    wxIViewCanvasSizeEvent oEvent;

    ProcessEvent( oEvent );
}

// Handle mouse events.
//
void wxIViewCanvas::OnMouse( wxMouseEvent& oEvent )
{
    if( m_nOptions & wxICANVAS_MOUSE_SCROLL )
    {
        // Process right mouse button events.
        if( oEvent.RightDown() )
        {
            // Change mouse cursor.
            SetCursor( wxCursor( wxCURSOR_HAND ) );

            // Remember former mouse position.
            m_oMousePosOld= wxPoint( oEvent.GetX(), oEvent.GetY() );
        }
        else if ( oEvent.RightUp() )
        {
            // Reset mouse cursor.
            SetCursor( wxNullCursor );
        }
        else if ( oEvent.Dragging() && oEvent.RightIsDown() )
        {
            // Get mouse position.
            wxPoint oMousePos( oEvent.GetX(), oEvent.GetY() );

            // Get the virtual position of upper left corner
            // (the view start).
            wxPoint nViewStart;

            GetViewStart( &nViewStart.x, &nViewStart.y );

            // Calculate scrolling parameters and scroll image.
            //
            // NOTE: This only works because scroll unit size ist set to
            //       1 pixel per horizontal/vertikal scroll unit in
            //       void wxIViewCanvas::SetBitmap( const wxBitmap &oBitmap ).
            //
            Scroll( nViewStart.x - ( oMousePos.x - m_oMousePosOld.x ),
                    nViewStart.y - ( oMousePos.y - m_oMousePosOld.y ) );

            // Save the mouse position for further use.
            m_oMousePosOld= oMousePos;
        }
    }
    if( m_nOptions & wxICANVAS_MOUSE_SELECTION )
    {
        // Process left mouse button events
        if( oEvent.LeftDown() )
        {
            // Clear a posible selection.
            ClearSelection();

            // Change mouse cursor.
            SetCursor( wxCursor( wxCURSOR_CROSS ) );

            // Remember former mouse position.
            m_oMousePosOld= wxPoint( oEvent.GetX(), oEvent.GetY() );

            // Save starting point of the mouse movement.
            m_oMousePosBegin= m_oMousePosOld;
        }
        else if ( oEvent.LeftUp() )
        {
            // Reset mouse cursor.
            SetCursor( wxNullCursor );

            // Get mouse position.
            wxPoint oMousePos( oEvent.GetX(), oEvent.GetY() );

            // Get the virtual position of upper left corner
            // (the view start).
            wxPoint nViewStart;

            GetViewStart( &nViewStart.x, &nViewStart.y );

            // Remember selection.
            m_oSelection= wxRect( nViewStart.x + m_oMousePosBegin.x, nViewStart.y + m_oMousePosBegin.y,
                                  oMousePos.x - m_oMousePosBegin.x + 1, oMousePos.y - m_oMousePosBegin.y + 1 );

            if( m_oSelection.width < 0 )
            {
                m_oSelection.x += m_oSelection.width;
                m_oSelection.width= -m_oSelection.width;
            }
            if( m_oSelection.height < 0 )
            {
                m_oSelection.y += m_oSelection.height;
                m_oSelection.height= -m_oSelection.height;
            }
            if( m_oSelection.x < 0 )
            {
                m_oSelection.width += m_oSelection.x;
                m_oSelection.x= 0;
            }
            if( m_oSelection.y < 0 )
            {
                m_oSelection.height += m_oSelection.y;
                m_oSelection.y= 0;
            }
        }
        else if ( oEvent.Dragging() && oEvent.LeftIsDown() )
        {
            // Get mouse position.
            wxPoint oMousePos( oEvent.GetX(), oEvent.GetY() );

            // ..
            wxClientDC oDc( this );

            oDc.SetPen( *wxBLACK_DASHED_PEN );
            oDc.SetBrush( *wxTRANSPARENT_BRUSH );
            oDc.SetLogicalFunction( wxINVERT );
            oDc.DrawRectangle( m_oMousePosBegin.x, m_oMousePosBegin.y,
                               m_oMousePosOld.x - m_oMousePosBegin.x + 1, m_oMousePosOld.y - m_oMousePosBegin.y + 1 );
            oDc.DrawRectangle( m_oMousePosBegin.x, m_oMousePosBegin.y,
                               oMousePos.x - m_oMousePosBegin.x + 1, oMousePos.y - m_oMousePosBegin.y + 1 );

            // Save the mouse position for further use.
            m_oMousePosOld= oMousePos;
        }
    }
}

// Repaints the image (using an internal bitmap).
//
void wxIViewCanvas::Paint( wxDC& oDc )
{
    if( oDc.Ok() )
    {
        // If there is a bitmap connected to this canvas window (re)paint it.
        if( m_oBitmap.Ok() )
        {
            // Repaint the bitmap.
            oDc.DrawBitmap( m_oBitmap, 0, 0 );

        }

        // If there is a selection repaint the surrounding rectangle.
        if( IsSelected() )
        {
            oDc.SetPen( *wxBLACK_DASHED_PEN );
            oDc.SetBrush( *wxTRANSPARENT_BRUSH );
            oDc.SetLogicalFunction( wxINVERT );
            oDc.DrawRectangle( m_oSelection.x, m_oSelection.y, m_oSelection.width, m_oSelection.height );
        }
    }
}

// Sets the internal bitmap.
//
void wxIViewCanvas::SetBitmap( const wxBitmap &oBitmap )
{
    // Save the (new) bitmap,...
    m_oBitmap= oBitmap;

    // ... adjust the scrollbars, ...
    SetScrollbars( 1, 1, m_oBitmap.GetWidth(), m_oBitmap.GetHeight() );

    // ... clear a posible selection, ...
    ClearSelection();

    // .. and refresh (repaint) the windown content.
    Refresh();
}
